/*
 * Gestures Security System interface
 * By Daniel Meja V.
 */

/*
 * PINS:
 * D2: PS/2 Clock (red wire)
 * D3: PS/2 Data (brown wire)
 * A0: Action button (Start/Stop password input)
 *
 */

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/delay.h>
#include "pin_access.h"
#include "uart.h"
#include "touchpad.h"
#include "gestures.h"


// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

// MACROS
#define ABSOLUTE_BYTE_COUNT 6
#define TPBUFFER_SIZE 20 // Needs to be bigger than ABSOLUTE_BYTE_COUNT
#define TPSCAN_TIME 50 // Check the touchpad every __ms
#define BUTTONCHECK_TIME 50 // Time to check the action button

#define ACTION_BUTTON_PORT PINA
#define ACTION_BUTTON_PORTDIR DDRA
#define ACTION_BUTTON_PIN PINA0

#define PW_MAX_SIZE 400

// TYPE DEFINITIONS
typedef enum{
	INTERFACE_IDLE,
	INTERFACE_RECEIVING_PW,
	INTERFACE_PW_ACCEPTED,
	INTERFACE_PW_DENIED
} interface_state_t;

typedef enum {
	BUTTON_IDLE,
	BUTTON_MAYBE_PRESSED,
	BUTTON_PRESSED,
	BUTTON_MAYBE_RELEASED,
	BUTTON_RELEASED
} button_state_t;

// PROTOTYPES
void enable_tp_ISR();
void disable_tp_ISR();

// GLOBAL VARIABLES
volatile byte tpdata[TPBUFFER_SIZE], newdata;
volatile unsigned int tpnumbytes, time_scan, time_button;
byte button_actioned, action_received;\
interface_state_t istate;
button_state_t bt_state;
unsigned int pw_size;
gesture_data_t pw[PW_MAX_SIZE];
byte timeout;

// Variables for testing
unsigned int pw_size_old;
gesture_data_t pw_old[PW_MAX_SIZE];
unsigned int pw_count;



/*
 * External interrupt ISR
 */
ISR(INT0_vect) {
	tpdata[tpnumbytes++] = tp_receive_data();
	//tp_inhibit(); // Already inhibited in tp_get_data
	if (tpnumbytes < ABSOLUTE_BYTE_COUNT)
		tp_disinhibit();
	else
		newdata = 1;

	SET(EIFR, INTF0); // Clears the flag (for interrupts that happened while in the ISR)
}

/*
 * Timer 0 Compare Match ISR
 */
ISR(TIMER0_COMPA_vect) {
	if (time_scan < TPSCAN_TIME) {
		time_scan++;
	}
	if (time_button < BUTTONCHECK_TIME) {
		time_button++;
	}
}

/*
 * Initializes the MCU and all the global variables
 */
void initialize() {
	// Rx Tx Debugging pins
	CLR(DDRD, PIND0);
	SET(DDRD, PIND1);

	// LED test pins
	DDRB = 0xFF;
	PORTB = 0xFF;

	CLR(ACTION_BUTTON_PORTDIR, ACTION_BUTTON_PIN);

	// Set timer 0 for 1ms base
    TIMSK0= (1<<OCIE0A); // Turn on Compare Match ISR
    OCR0A = 249; // Compare register to 250 time ticks
    TCCR0B= 3; // Prescalar of 64
    TCCR0A= (1<<WGM01); // Clear on match

	newdata = 0;
	time_button = 0;
	time_scan = 0;
	pw_size = 0;
	timeout = 0;
	pw_count = 0;
	istate = INTERFACE_IDLE;
	bt_state = BUTTON_IDLE;
	uart_init();
	stdout = stdin = stderr = &uart_str;
	sei();
	fprintf(stdout,"Hello\n");
}

/*
 * Checks if the connected device is a valid
 * touchpad, and returns -1 if it's not.
 * Otherwise, initializes the touchpad in absolute mode,
 * and enables ISR for data reception.
 */
int init_touchpad() {
	int result;
	fprintf(stdout, "\n\n\n\n\n\n\n\n");
	tp_init_ports();
	tp_disinhibit();
	if ((result = tp_check_valid_device()) == -1) {
		fprintf(stdout, "Not a valid touchpad\nResult was %x\n", result&0xFF);
		return -1;
	}
	fprintf(stdout, "Valid touchpad!\n");
	tp_disable_packet_transmission();
	tp_set_mode_byte(1, 0, 0, 0, 0, 1); // Absolute mode, with "W" enabled
	tp_enable_packet_transmission();
	//tp_inhibit();
	enable_tp_ISR();
	return 0;
}

/*
 * Enables the external interrupt ISR for
 * data reception from the touchpad.
 */
void enable_tp_ISR() {
	CLR(EIMSK, INT0); // So that the change in EICRA does not trigger an interrupt
	EICRA = 2; // Falling Edge
	SET(EIFR, INTF0); // Clears the flag
	SET(EIMSK, INT0); // Enable external interrupt in PIN D2

}

/*
 * Disables the external interrupt ISR for
 * data reception from the touchpad.
 */
void disable_tp_ISR() {
	CLR(EIMSK, INT0);
	EICRA = 0;
}

/*
 * Initializes the reception of data
 * from the touchpad
 */
void init_tp_rx() {
	newdata = 0;
	tp_disinhibit();
	tpnumbytes = 0;
}

/*
 * Scans and returns the data obtained from the touchpad.
 * If no new data, the returned gesture_data_t variable
 * has all fields as zero.
 */
gesture_data_t scan_tp() {
	tp_abs_packet_t packet;
	gesture_data_t gdata;
	if (newdata) {
		packet = tp_parse_absolute_packet((byte *)tpdata);
		return touchpad2gesture_data(packet);
	}
	gdata.xpos = 0;
	gdata.ypos = 0;
	gdata.zpress = 0;
	return gdata;
}

/*
 * Initialization for receiving a new password
 */
void init_pw_rx() {
	pw_size = 0;
}

/*
 * Updates the password with the received data.
 * Returns -1 if the size limit was exceeded.
 */
int update_pw(gesture_data_t data) {
	if (pw_size == PW_MAX_SIZE - 1)
		return -1;
	pw[pw_size++] = data;
	return 0;
}
/*
 * DEBUG function that prints the pw
 */
void print_pw() {
	int i;
	fprintf(stdout, "\n\nThe password is:\n\n\n");
	for (i = 0; i < pw_size; i++) {
		fprintf(stdout, "PW UNIT %d:\nx:%d\ny:%d\nz:%d\n\n",
				i, pw[i].xpos, pw[i].ypos, pw[i].zpress);
	}
	fprintf(stdout, "\n\n\n");
}

/*
 * Checks if the entered password
 * is valid
 */
int valid_pw() {
	return gesture_check_match(pw_old, pw_size_old, pw, pw_size);
}

/*
 * FSM that checks if the action button was
 * pressed. If it was, then button_actioned is set to 1.
 * Once the action is detected, the FSM waits until
 * action_received is 1 to go back to checking the button.
 */
void check_button() {
	switch(bt_state) {
	case BUTTON_IDLE:
		if(READ(ACTION_BUTTON_PORT, ACTION_BUTTON_PIN) == 0) {
			bt_state = BUTTON_MAYBE_PRESSED;
		}
		break;
	case BUTTON_MAYBE_PRESSED:
		if(READ(ACTION_BUTTON_PORT, ACTION_BUTTON_PIN) == 0) {
			bt_state = BUTTON_PRESSED;
			button_actioned = 1;
		} else
			bt_state = BUTTON_IDLE;
		break;
	case BUTTON_PRESSED:
		if(READ(ACTION_BUTTON_PORT, ACTION_BUTTON_PIN) == 1) {
			bt_state = BUTTON_MAYBE_RELEASED;
		}
		break;
	case BUTTON_MAYBE_RELEASED:
		if(READ(ACTION_BUTTON_PORT, ACTION_BUTTON_PIN) == 1) {
			bt_state = BUTTON_RELEASED;
		} else {
			bt_state = BUTTON_PRESSED;
		}
		break;
	case BUTTON_RELEASED:
		if (action_received) {
			action_received = 0;
			bt_state = BUTTON_IDLE;
		}
		break;
	}
}

/*
 * Updates the main interface FSM.
 */
void update_interface_state() {
	int i; // Used for testing
	switch (istate) {
	case INTERFACE_IDLE:
		if (button_actioned) {
			button_actioned = 0;
			action_received = 1;
			istate = INTERFACE_RECEIVING_PW;
			init_tp_rx();
			init_pw_rx();

		}
		break;
	case INTERFACE_RECEIVING_PW:
		if (timeout) {
			timeout = 0;
			istate = INTERFACE_PW_DENIED;
		}
		else if (button_actioned) {
			button_actioned = 0;
			action_received = 1;
			istate = INTERFACE_IDLE;
			//tp_inhibit();  // Would cause an interrupt and get stuck.
							 // The touchpad is inhibited in the interrupt anyway.

			/*if (valid_pw())
				istate = INTERFACE_PW_ACCEPTED;
			else
				istate = INTERFACE_PW_DENIED;*/

			// For testing purposes
			if (pw_count == 0) {
				pw_size_old = pw_size;
				for (i = 0; i< pw_size; i++) {
					pw_old[i] = pw[i];
				}
				istate = INTERFACE_PW_ACCEPTED;
				pw_count++;
			} else {
				if(valid_pw())
					istate = INTERFACE_PW_ACCEPTED;
				else
					istate = INTERFACE_PW_DENIED;
			}

		}
		break;
	case INTERFACE_PW_ACCEPTED:
		print_pw();
		istate = INTERFACE_IDLE;
		break;
	case INTERFACE_PW_DENIED:
		// TODO signal_pw_denied();
		fprintf(stdout, "Invalid Password\n\n");
		istate = INTERFACE_IDLE;
		break;
	default:
		break;
	}
}

int main(void) {
	initialize();
	if (init_touchpad() == -1) // Invalid touchpad
		return -1;
	while (1) {
		if (time_button >= BUTTONCHECK_TIME){
			check_button();
			time_button = 0;
		}

		update_interface_state();

		if (time_scan >= TPSCAN_TIME) {
			if (istate == INTERFACE_RECEIVING_PW) {
				//fprintf(stdout, "Receiving PW\n");

				if (update_pw(scan_tp()) == -1) {
					// Size limit reached
					timeout = 1;
				} else
					init_tp_rx(); // Re-enable reception from the touchpad
			}
			time_scan = 0;
		}

	}

}


