/*
 * Gesture Manager
 * Implementation file.
 * By Daniel Meja V.
 */


#include "gestures.h"
#include "string.h"
#include <stdio.h> // DEBUGGING, REMOVE!
#include "uart.h" // DEBUGGING, REMOVE!
#include "strokes.h"
#include <stdlib.h>

#define MAX_INITIAL_POS_DIFF 6
#define DELTAS_ERROR_CONSTANT 3
#define MAX_STROKE_SIZES_DIFF 40 // Max diff (in % wrt the smaller) of stroke sizes
#define MAX_WAIT_TIME_DIFF 80 // (In % wrt the smaller)


// UART file descriptor FOR DEBUGGIN PURPOSES! REMOVE!
// putchar and getchar are in uart.c
extern FILE uart_str;

// PROTOTYPES
byte normalize_xpos_value(unsigned int value);
byte normalize_ypos_value(unsigned int value);
void remove_trailing_zeros(gesture_raw_unit_t *src, gesture_raw_unit_t *dst,
						   int size, int *new_size);
int strokes_match(stroke_t *strokes1, int num_strokes1,
				  stroke_t *strokes2, int num_strokes2);
inline int get_min(int x, int y);
inline int get_abs(int x);
void print_gesture(gesture_raw_unit_t *gest, int size); // FOR DEBUGGING


// PUBLIC FUNCTIONS

/*
 * Checks if the two given gestures match
 */
int gesture_check_match(gesture_raw_unit_t *g1, unsigned int g1_size,
						gesture_raw_unit_t *g2, unsigned int g2_size) {
	int g1_newsize, g2_newsize, result;
	gesture_raw_unit_t *g1_new;
	gesture_raw_unit_t *g2_new;
	stroke_t *strokes1;
	stroke_t *strokes2;
	int num_strokes1, num_strokes2;

	// Remove zeroes at the beginning and at the end
	g1_new = (void *)malloc(g1_size*sizeof(gesture_raw_unit_t));
	remove_trailing_zeros(g1, g1_new, g1_size, &g1_newsize);
	//fprintf(stdout, "\nGesture 1 raw:\n");
	//print_gesture(g1_new, g1_newsize);
	// Create strokes from the data
	strokes1 = strokes_create(g1_new, g1_newsize, &num_strokes1);
	free(g1_new);

	// Remove zeroes at the beginning and at the end
	g2_new = (void *)malloc(g2_size*sizeof(gesture_raw_unit_t));
	remove_trailing_zeros(g2, g2_new, g2_size, &g2_newsize);
	//fprintf(stdout, "\nGesture 2 raw:\n");
	//print_gesture(g2_new, g2_newsize);
	// Create strokes from the data
	strokes2 = strokes_create(g2_new, g2_newsize, &num_strokes2);
	free(g2_new);

	// Compare the strokes
	result =  strokes_match(strokes1, num_strokes1, strokes2, num_strokes2);
	strokes_free(strokes1, num_strokes1);
	strokes_free(strokes2, num_strokes2);

	return result;

}

/*
 * Transforms the touchpad packet into the data
 * representation that is used for gesture recognition.
 */
gesture_raw_unit_t touchpad2gesture_raw_unit(tp_abs_packet_t packet) {
	gesture_raw_unit_t gdata;
	gdata.xpos = normalize_xpos_value(packet.xpos);
	gdata.ypos = normalize_ypos_value(packet.ypos);
	gdata.zpress = packet.zpress;
	return gdata;
}

/*
 * Returns 1, if the data represents a valid touch.
 * i.e zpress higher than the threshold
 */
int valid_touch(gesture_raw_unit_t data) {
	return data.zpress >= ZPRESS_TOUCH_THRESHOLD;
}



// PRIVATE FUNCTIONS

int strokes_match(stroke_t *strokes1, int num_strokes1,
				  stroke_t *strokes2, int num_strokes2) {

	// TODO: Consider changing the % to be with respect to the model to match
	int i, j, deltas_error, wait_error, stroke_sizes_diff;
	int min_delta_size;
	// int cumul_error;
	stroke_t stroke1, stroke2;

	if (num_strokes1 != num_strokes2) {
		fprintf(stdout, "Num strokes1: %d, Num strokes2: %d\n\n", num_strokes1, num_strokes2);
		return 0;
	}

	for (i = 0; i < num_strokes1; i++) {
		//fprintf(stdout, "****** Stroke number %d ******\n\n", i);
		stroke1 = *(strokes1 + i);
		stroke2 = *(strokes2 + i);
		// First compare initial position

		if (get_abs(stroke1.x_initial - stroke2.x_initial) +
		    get_abs(stroke1.y_initial - stroke2.y_initial) > MAX_INITIAL_POS_DIFF) {
			fprintf(stdout, "Stroke %d\n", i);
			fprintf(stdout, "stroke1 x initial: %d, y initial: %d\n", stroke1.x_initial, stroke1.y_initial);
			fprintf(stdout, "stroke2 x initial: %d, y initial: %d\n", stroke2.x_initial, stroke2.y_initial);
			return 0;
		}


		// Second compare the deltas
		min_delta_size = get_min(list_size(stroke1.x_deltas), list_size(stroke2.x_deltas));
		//fprintf(stdout, "\nStroke1 size: %d, Stroke2 size: %d\n\n", list_size(stroke1.x_deltas),
			//	list_size(stroke2.x_deltas));

		// If the difference in size is too big, reject match right away
		stroke_sizes_diff = (100*get_abs(list_size(stroke1.x_deltas) -
							 list_size(stroke2.x_deltas))) / min_delta_size;

		if (stroke_sizes_diff > MAX_STROKE_SIZES_DIFF ) {
			fprintf(stdout, "Stroke %d\n", i);
			fprintf(stdout, "size diff error: %d\n", stroke_sizes_diff);
			return 0;
		}


		deltas_error = 0;
		//fprintf(stdout, "Deltas:\n");
		for (j = 0; j < min_delta_size; j++) {
			/*fprintf(stdout, "\nPos %d\n\n", j);
			fprintf(stdout, "Strk1 dx: %d\n", list_get(stroke1.x_deltas, j));
			fprintf(stdout, "Strk1 dy: %d\n\n", list_get(stroke1.y_deltas, j));
			fprintf(stdout, "Strk2 dx: %d\n", list_get(stroke2.x_deltas, j));
			fprintf(stdout, "Strk2 dy: %d\n", list_get(stroke2.y_deltas, j));*/
			deltas_error += get_abs((int)list_get(stroke1.x_deltas, j)-(int)list_get(stroke2.x_deltas, j)) +
							get_abs((int)list_get(stroke1.y_deltas, j)-(int)list_get(stroke2.y_deltas, j));
		}

		if (deltas_error > DELTAS_ERROR_CONSTANT*min_delta_size) {
			fprintf(stdout, "Stroke %d\n", i);
			fprintf(stdout, "\ndeltas error: %d\n", deltas_error);
			return 0;
		}

		// Third compare the wait periods
		wait_error = 100*get_abs(stroke1.wait_time - stroke2.wait_time) /
							 get_min(stroke1.wait_time, stroke2.wait_time); // TODO Set a min abs valule(becasue very hard for short wait periods

		if (wait_error > MAX_WAIT_TIME_DIFF) {
			fprintf(stdout, "Stroke %d\n", i);
			fprintf(stdout, "\nwait error: %d\n", wait_error);
			fprintf(stdout, stroke2.wait_time > stroke1.wait_time ?
					"Stroke 2 too long" : "Stroke 2 too short\n\n");
			return 0;
		}

	}
	// TODO  Check cumulative error

	return 1;

}

/*
 * Normalizes the touchpad x position value
 * into a discrete representation used
 * for gestures recognition
 */
byte normalize_xpos_value(unsigned int value) {
	byte xpos;
	int step = (TP_HIGHER_BEZEL_XPOS - TP_LOWER_BEZEL_XPOS)/ GRID_SIZE_X;
	value = value < TP_LOWER_BEZEL_XPOS ? 0 : value - TP_LOWER_BEZEL_XPOS;
	xpos = value/step;
	return xpos < GRID_SIZE_X ? xpos : GRID_SIZE_X -1;
}

/*
 * Normalizes the touchpad y position value
 * into a discrete representation used
 * for gestures recognition
 */
byte normalize_ypos_value(unsigned int value) {
	byte ypos;
	int step = (TP_HIGHER_BEZEL_YPOS - TP_LOWER_BEZEL_YPOS)/ GRID_SIZE_Y;
	value = value < TP_LOWER_BEZEL_YPOS ? 0 : value - TP_LOWER_BEZEL_YPOS;
	ypos = value/step;
	return ypos < GRID_SIZE_Y ? ypos : GRID_SIZE_Y -1;
}

/*
 * Removes the initial and final zeroes that come from the
 * time the user takes from/to pressing the bottom
 */
void remove_trailing_zeros(gesture_raw_unit_t *src, gesture_raw_unit_t *dst, int size, int *new_size) {
	int i, initialz_done = 0, num_initialz = 0, num_finalz = 0;
	*new_size = 0;
	// Remove initial zeros
	for (i = 0; i < size; i++) {
		if (valid_touch(*(src+i)) && initialz_done == 0) {
			initialz_done = 1;
			num_initialz = i;
		}
		if (initialz_done == 1) {
			memcpy(dst+i-num_initialz, src+i, sizeof(gesture_raw_unit_t));
		}
	}
	// Count the number of final zeros
	for (i = size-num_initialz-1; valid_touch(*(dst+i)) == 0; i--) {
		num_finalz++;
	}
	*new_size = size - num_initialz - num_finalz;
}


/*
 * Returns the absolute value of x
 */
inline int get_abs(int x) {
	return x >= 0 ? x : -x;
}

/*
 * Returns the minimum between x and y
 */
inline int get_min(int x, int y) {
	return x < y ? x : y;
}

/*
 * Prints the whole gesture
 * USED FOR DEBUGGING, REMOVE
 */
void print_gesture(gesture_raw_unit_t *gest, int size) {
	int i;
	fprintf(stdout, "\n\nThe gesture is:\n\n\n");
	for (i = 0; i < size; i++) {
		fprintf(stdout, "GESTURE UNIT %d:\nx:%d\ny:%d\nz:%d\n\n",
				i, gest[i].xpos,gest[i].ypos, gest[i].zpress);
	}
	fprintf(stdout, "\n\n\n");
}


