/*
 * PS/2 Controller
 * Implementation file.
 * By Daniel Meja V.
 */

#include "ps2.h"
#include "pin_access.h"
#include <util/delay.h>


// PROTOTYPES

byte read_clock();
byte read_data();
void set_clock(byte value);
void set_data(byte value);
void wait_us(byte us);
void wait_clock(byte state);
void error();


/** PUBLIC FUNCTIONS **/

/*
 * Returns a byte of data received from
 * a PS/2 device.
 * Waits for the clock to go low.
 */
byte ps2_get() {

	// Start bit
	set_clock(1); // In case the device was disabled (Clock pulled down)
	wait_clock(1);
	wait_clock(0);
	return ps2_receive();
}

/*
 * Returns a byte of data received from
 * a PS/2 device, without waiting for the clock
 * pin to go low. This should happen before calling
 * the function.
 */
byte ps2_receive() {
	byte i, bit, value = 0, p = 0;

	wait_clock(1); // Start bit over

	// Data
	for (i = 0; i < 8; i++) {
		wait_clock(0);
		bit = read_data();
		value = value + (bit << i);
		p = p + bit;
		wait_clock(1);

	}

	// Parity
	wait_clock(0);
	p = p + read_data();
	if ((p & 0x01) == 0)
		error(); // Parity error
	wait_clock(1);

	// Stop bit
	wait_clock(0);
	if (read_data() == 0)
		error(); // Framing error
	//SET(PS2ERROR_PORT, PS2ERR0R_PIN);
	set_clock(0);
	wait_us(50); // To make sure another immediate call to get doesn't get a wrong start bit
	return value;
}

/*
 * Sends a byte of data to
 * a PS/2 device
 */
void ps2_send(byte value) {
	byte i, ack, p = 1;//ones = 0;

	// Inhibit, if necessary
	set_clock(0);
	wait_us(100);

	while (1) {
		// Request to send
		set_data(0);
		set_clock(1);
		wait_clock(1); // Wait until it actually is 1 (previous instruction simply released it)

		// Data
		for (i = 0; i < 8; i++) {
			wait_clock(0);
			set_data(value & 0x01);
			//ones = ones + (value & 0x01); // Accumulate 1's for parity
			p = p + value;
			wait_clock(1);
			value = value >> 1;
		}

		// Parity
		wait_clock(0);
		//set_data(~(ones & 0x01));
		set_data(p & 0x01);
		wait_clock(1);

		// Stop bit
		wait_clock(0);
		set_data(1);
		wait_clock(1);

		// Line control bit
		wait_clock(0);
		if (read_data() == 1)
			error(); // Missing line control bit
		wait_clock(1);

		// Check ACK
		ack = ps2_get();
		if (ack != PS2COMMAND_ACK) {
			if (ack == PS2COMMAND_RESEND)
				CLR(PORTB, PINB4);
			else if (ack == PS2COMMAND_ERROR) {
				CLR(PORTB, PINB5);
				error();
				break;
			}
		} else {
			CLR(PORTB, PINB3);
			break;
		}
	}
}

/*
 * Inhibits PS/2 device by
 * pulling the clock down
 */
void ps2_inhibit_device() {
	set_clock(0);
}

/*
 * Disihibits PS/2 device
 */
void ps2_disinhibit_device() {
	set_clock(1);
	wait_clock(1); // TODO: Make sure it's necessary
}

/*
 * Initializes the ports used to
 * access the P2/2 device
 */
void ps2_init_ports() {
	CLR(PS2CLOCK_PORTDIR, PS2CLOCK_PIN); // Clock pin as input
	CLR(PS2DATA_PORTDIR, PS2DATA_PIN); // Data pin as input
}



/** PRIVATE FUNCTIONS **/

/*
 * Returns the value of the clock pin.
 */
byte read_clock() {
	return READ(PS2CLOCK_PORT_READ, PS2CLOCK_PIN);
}

/*
 * Returns the value of the data pin.
 */
byte read_data() {
	return READ(PS2DATA_PORT_READ, PS2DATA_PIN);
}

/*
 * Sets the value of the clock pin.
 */
void set_clock(byte value) {
	if (value) {
		CLR(PS2CLOCK_PORTDIR, PS2CLOCK_PIN);
	} else {
		SET(PS2CLOCK_PORTDIR, PS2CLOCK_PIN);
		CLR(PS2CLOCK_PORT_WRITE, PS2CLOCK_PIN);
	}
}

/*
 * Sets the value of the data pin.
 */
void set_data(byte value) {
	if (value) {
		CLR(PS2DATA_PORTDIR, PS2DATA_PIN);
	} else {
		SET(PS2DATA_PORTDIR, PS2DATA_PIN);
		CLR(PS2DATA_PORT_WRITE, PS2DATA_PIN);
	}
}

/*
 * Delay of us microseconds
 */
void wait_us(byte us) {
	_delay_us(us);
}

/*
 * Waits until the clock has the value
 * of state
 */
void wait_clock(byte state) {
	while(read_clock() != state);
}


/*
 * Reports an error
 */
void error() {
	CLR(PS2ERROR_PORT, PS2ERR0R_PIN);

}
